#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from unittest import TestCase
from unittest.mock import patch, Mock

from kios import config
from kios.db import App, PersistenceManager
from kios.exception import UnsupportedAppPlatformError
from kios import factory
from .helper import patch_config_app_platform, temp_db_real_session_factory
from kios.parser import WindowsNetstatPortDataParser, PortCollector
from kios.system import WindowsNetstatDataSource


class SessionFactoryTestCase(TestCase):

    def test_session_factory_lifecycle(self):
        with temp_db_real_session_factory() as sf:
            s = sf()
            try:
                self.assertEqual(s.query(App).count(), 0)
                r = s.execute('PRAGMA foreign_keys')
                fk, = r.fetchone()
                self.assertTrue(bool(fk))
                r = s.execute('PRAGMA busy_timeout')
                bt, = r.fetchone()
                self.assertEqual(bt, config.db_busy_timeout)
            finally:
                s.close()
            sf.dispose()
            with self.assertRaises(RuntimeError):
                sf()

    @patch('kios.config.db_busy_timeout', new='test')
    def test_session_factory_lifecycle2(self):
        with self.assertRaises(RuntimeError):
            with temp_db_real_session_factory():
                pass

    @patch('kios.config.db_file', new=None)
    def test_missing_db_file_path(self):
        with self.assertRaises(UnsupportedAppPlatformError):
            factory.SessionFactory()


class DataObjectTestCase(TestCase):

    @patch_config_app_platform
    def test_creating_data_source(self):
        self.assertTrue(isinstance(factory.data_source(), WindowsNetstatDataSource))

    def test_creating_data_source2(self):
        with self.assertRaises(UnsupportedAppPlatformError):
            factory.data_source()

    @patch_config_app_platform
    def test_creating_data_parser(self):
        self.assertTrue(isinstance(factory.data_parser(Mock(spec_set=PortCollector)), WindowsNetstatPortDataParser))

    def test_creating_data_parser2(self):
        with self.assertRaises(UnsupportedAppPlatformError):
            factory.data_parser(Mock(spec_set=PortCollector))

    def test_creating_persistence_manager(self):
        self.assertTrue(isinstance(factory.persistence_manager(), PersistenceManager))



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from .data import NetworkProtocol, TransportProtocol


class OperationControl(Protocol):
    interrupted: bool

    @abstractmethod
    def feedback_pass_done(self):
        raise NotImplementedError


class Feedback(Protocol):

    @abstractmethod
    def app_entry(self, name: str):
        raise NotImplementedError

    @abstractmethod
    def port_entry(self, np: NetworkProtocol, tp: TransportProtocol, number: int):
        raise NotImplementedError

    @abstractmethod
    def executable_entry(self, name: str):
        raise NotImplementedError



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from unittest import TestCase
from unittest.mock import Mock, call

from kios.data import NetworkProtocol, TransportProtocol, ConnectionState, NETWORK_PROTOCOL, TRANSPORT_PROTOCOL, \
    PORT_NUMBER, EXECUTABLE_NAME
from kios.parser import PortCollector, WindowsNetstatPortDataParser
from kios.exception import UnexpectedLineError
from .helper import get_test_data, open_test_file, patch_config_app_platform


class PortCollectorTestCase(TestCase):

    def setUp(self) -> None:
        self.pc = PortCollector()

    def test_callback_capability(self):
        self.pc('test.exe', NetworkProtocol.IPv4, '0.0.0.0', TransportProtocol.TCP, 1, ConnectionState.LISTENING)
        self.pc(None, NetworkProtocol.IPv6, '::', TransportProtocol.TCP, 2, ConnectionState.ESTABLISHED)
        self.pc('test2.exe', NetworkProtocol.IPv6, '2001:db8::1000', TransportProtocol.UDP, 3, None)
        self.pc('test.exe', NetworkProtocol.IPv4, '127.0.0.1', TransportProtocol.TCP, 1, ConnectionState.LISTENING)
        self.pc(None, NetworkProtocol.IPv6, '::1', TransportProtocol.TCP, 2, ConnectionState.LISTENING)
        self.assertSequenceEqual(self.pc.port_data, [{EXECUTABLE_NAME: 'test.exe',
                                                      NETWORK_PROTOCOL: NetworkProtocol.IPv4,
                                                      TRANSPORT_PROTOCOL: TransportProtocol.TCP, PORT_NUMBER: 1},
                                                     {EXECUTABLE_NAME: 'test2.exe',
                                                      NETWORK_PROTOCOL: NetworkProtocol.IPv6,
                                                      TRANSPORT_PROTOCOL: TransportProtocol.UDP, PORT_NUMBER: 3}])
        self.pc.reset()
        self.assertEqual(len(self.pc.port_data), 0)


@patch_config_app_platform
class WindowsNetstatPortDataParserTestCase(TestCase):
    expected_calls = [call('app1.exe', NetworkProtocol.IPv4, '0.0.0.0', TransportProtocol.TCP, 1,
                           ConnectionState.LISTENING),
                      call(None, NetworkProtocol.IPv4, '0.0.0.0', TransportProtocol.TCP, 2,
                           ConnectionState.LISTENING),
                      call('app2.exe', NetworkProtocol.IPv4, '0.0.0.0', TransportProtocol.TCP, 3,
                           ConnectionState.ESTABLISHED),
                      call(None, NetworkProtocol.IPv4, '0.0.0.0', TransportProtocol.TCP, 4,
                           ConnectionState.ESTABLISHED),
                      call('app3.exe', NetworkProtocol.IPv6, '::', TransportProtocol.TCP, 15,
                           ConnectionState.CLOSE_WAIT),
                      call('app1.exe', NetworkProtocol.IPv6, '::', TransportProtocol.TCP, 1,
                           ConnectionState.LISTENING),
                      call(None, NetworkProtocol.IPv6, '::', TransportProtocol.TCP, 2,
                           ConnectionState.LISTENING),
                      call('app2.exe', NetworkProtocol.IPv4, '0.0.0.0', TransportProtocol.UDP, 1, None),
                      call(None, NetworkProtocol.IPv4, '1.2.3.4', TransportProtocol.UDP, 2, None),
                      call('app1.exe', NetworkProtocol.IPv6, 'fe80::a1:b2:c3:d4', TransportProtocol.UDP, 1, None),
                      call(None, NetworkProtocol.IPv6, 'fe80::a1:b2:c3:d4', TransportProtocol.UDP, 2, None)]

    def setUp(self) -> None:
        self.pc = Mock(spec_set=PortCollector)()
        self.dp = WindowsNetstatPortDataParser(self.pc)

    def test_data_parsing(self):
        data = get_test_data('win_netstat.txt')

        self.dp.extract_data(data)
        self.assertEqual(self.pc.call_count, 11)
        self.pc.assert_has_calls(self.expected_calls, any_order=True)
        # verify that parser can do multiple parses
        self.pc.reset_mock()
        self.dp.extract_data(data)
        self.assertEqual(self.pc.call_count, 11)
        self.pc.assert_has_calls(self.expected_calls, any_order=True)

    def test_line_stream_parsing(self):
        with open_test_file('win_netstat.txt') as data_file:
            with self.dp.line_stream_consumer() as consume:
                for line_no, line in enumerate(data_file):
                    consume(line, line_no)
        self.assertEqual(self.pc.call_count, 11)
        self.pc.assert_has_calls(self.expected_calls, any_order=True)

    def test_unexpected_data_handling(self):
        data = get_test_data('win_netstat3.txt')
        with self.assertRaises(UnexpectedLineError) as c:
            self.dp.extract_data(data)
        self.assertEqual(c.exception.line_no, 6)
        self.assertEqual(c.exception.line, ' [app3.exe]')

    def test_unexpected_data_handling2(self):
        with self.assertRaises(UnexpectedLineError) as c:
            with open_test_file('win_netstat3.txt') as data_file:
                with self.dp.line_stream_consumer() as consume:
                    for line_no, line in enumerate(data_file):
                        consume(line, line_no + 1)
        self.assertEqual(c.exception.line_no, 6)
        self.assertEqual(c.exception.line, ' [app3.exe]')



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from unittest import TestCase
from unittest.mock import patch, call

from kios import config
from kios.data import EXECUTABLE_NAME, NETWORK_PROTOCOL, TRANSPORT_PROTOCOL, PORT_NUMBER, NetworkProtocol, \
    TransportProtocol
from kios.exception import DoRollback, UnexpectedLineError
from kios.operation import find_and_save_ports, save_ports_from_data_file
from .helper import get_test_data, TestOperationControl, get_test_file_path, patch_config_app_platform

_data_se = [get_test_data('win_netstat.txt'), get_test_data('win_netstat2.txt')]


path_config_data_file = patch('kios.config.data_file', new=get_test_file_path('win_netstat.txt'))

part1 = [{EXECUTABLE_NAME: 'app1.exe', NETWORK_PROTOCOL: NetworkProtocol.IPv4,
          TRANSPORT_PROTOCOL: TransportProtocol.TCP, PORT_NUMBER: 1},
         {EXECUTABLE_NAME: None, NETWORK_PROTOCOL: NetworkProtocol.IPv4,
          TRANSPORT_PROTOCOL: TransportProtocol.TCP, PORT_NUMBER: 2},
         {EXECUTABLE_NAME: 'app1.exe', NETWORK_PROTOCOL: NetworkProtocol.IPv6,
          TRANSPORT_PROTOCOL: TransportProtocol.TCP, PORT_NUMBER: 1}]
part2 = [{EXECUTABLE_NAME: None, NETWORK_PROTOCOL: NetworkProtocol.IPv6,
          TRANSPORT_PROTOCOL: TransportProtocol.TCP, PORT_NUMBER: 2},
         {EXECUTABLE_NAME: 'app2.exe', NETWORK_PROTOCOL: NetworkProtocol.IPv4,
          TRANSPORT_PROTOCOL: TransportProtocol.UDP, PORT_NUMBER: 1},
         {EXECUTABLE_NAME: None, NETWORK_PROTOCOL: NetworkProtocol.IPv4,
          TRANSPORT_PROTOCOL: TransportProtocol.UDP, PORT_NUMBER: 2}]
part3 = [{EXECUTABLE_NAME: 'app1.exe', NETWORK_PROTOCOL: NetworkProtocol.IPv6,
          TRANSPORT_PROTOCOL: TransportProtocol.UDP, PORT_NUMBER: 1},
         {EXECUTABLE_NAME: None, NETWORK_PROTOCOL: NetworkProtocol.IPv6,
          TRANSPORT_PROTOCOL: TransportProtocol.UDP, PORT_NUMBER: 2}]
port_data1 = [*part1, *part2, *part3]
port_data2 = [{EXECUTABLE_NAME: 'app3.exe', NETWORK_PROTOCOL: NetworkProtocol.IPv6,
               TRANSPORT_PROTOCOL: TransportProtocol.UDP, PORT_NUMBER: 3}]


@patch_config_app_platform
@patch('kios.factory.persistence_manager', autospec=True)
class OperationsTestCase(TestCase):

    @patch('kios.operation.time', autospec=True, side_effect=[0, 1, 6, 10, 0, 1, 6, 10])
    @patch('kios.operation.sleep', autospec=True)
    @patch('kios.factory.data_source', autospec=True, **{'return_value.get_port_data.side_effect': _data_se})
    def test_find_and_save_ports_behavior(self, data_source, sleep, time, persistence_manager):
        find_and_save_ports(TestOperationControl())
        sleep.assert_called_once_with(config.port_scan_wake_up_interval)
        self.assertEqual(persistence_manager.return_value.save_port_data.call_count, 2)
        persistence_manager.return_value.save_port_data.assert_has_calls([call(port_data1, commit=True),
                                                                          call(port_data2, commit=True)])
        self.assertEqual(data_source.return_value.get_port_data.call_count, 2)

    @path_config_data_file
    def test_save_ports_from_data_file_behavior(self, persistence_manager):
        save_ports_from_data_file(TestOperationControl())
        self.assertEqual(persistence_manager.return_value.save_port_data.call_count, 1)
        persistence_manager.return_value.save_port_data.assert_has_calls([call(port_data1)])

    @path_config_data_file
    @patch('kios.config.import_batch_size', new=3)
    def test_save_ports_from_data_file_behavior2(self, persistence_manager):
        save_ports_from_data_file(TestOperationControl(3))
        self.assertEqual(persistence_manager.return_value.save_port_data.call_count, 3)
        persistence_manager.return_value.save_port_data.assert_has_calls([call(part1), call(part2), call(part3)])

    @path_config_data_file
    @patch('kios.config.import_batch_size', new=3)
    def test_save_ports_from_data_file_behavior3(self, persistence_manager):
        with self.assertRaises(DoRollback):
            save_ports_from_data_file(TestOperationControl(2))
        self.assertEqual(persistence_manager.return_value.save_port_data.call_count, 2)
        persistence_manager.return_value.save_port_data.assert_has_calls([call(part1), call(part2)])

    @patch('kios.config.data_file', new=None)
    def test_save_ports_from_data_file_behaviour4(self, persistence_manager):
        with self.assertRaises(RuntimeError):
            save_ports_from_data_file(TestOperationControl())
        persistence_manager.return_value.save_port_data.assert_not_called()

    @patch('kios.config.data_file', new=get_test_file_path('win_netstat3.txt'))
    def test_save_ports_from_data_file_behaviour5(self, persistence_manager):
        with self.assertRaises(UnexpectedLineError) as c:
            save_ports_from_data_file(TestOperationControl())
        persistence_manager.return_value.save_port_data.assert_not_called()
        self.assertEqual(c.exception.line_no, 6)
        self.assertEqual(c.exception.line, ' [app3.exe]')



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING
from unittest.mock import patch

from sqlalchemy.event import listen
from sqlalchemy.orm import sessionmaker

from kios.cli import ConsoleOperationControl
from kios.data import Application, Platform
from kios import config
from kios import factory
from kios.protocol import OperationControl
from .singleton import engine

if TYPE_CHECKING:
    from sqlalchemy.orm import Session


def get_test_data(file_name: str):
    with open_test_file(file_name) as f:
        return f.read()


def open_test_file(file_name: str):
    return open(get_test_file_path(file_name), 'rt')


def get_test_file_path(file_name: str):
    return Path(__file__).parent / file_name


def _restart_nested_transaction(s, tr):
    if tr.nested and tr.parent is not None and not tr.parent.nested:
        s.expire_all()
        s.begin_nested()


path_config_intervals = patch.multiple(config, port_scan_interval=0, port_scan_wake_up_interval=0)
patch_config_app_platform = patch.multiple(config, app=Application.NETSTAT, platform=Platform.WINDOWS)


class TestOperationControl(OperationControl):

    def __init__(self, iterations: int = 2):
        self.interrupted = False
        self._current = 0
        self._max = iterations

    def feedback_pass_done(self):
        self._current += 1
        if self._current >= self._max:
            self.interrupted = True


inside_test_operation_control = ContextVar('inside_test_operation_control', default=None)


class TestConsoleOperationControl(ConsoleOperationControl):

    def __enter__(self):
        self._t = inside_test_operation_control.set(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        inside_test_operation_control.reset(self._t)


def inside_test_console_operation_control(oc):
    stored_oc = inside_test_operation_control.get()
    if stored_oc is None or stored_oc is not oc:
        raise AssertionError('Wrong or missing test console operation control context')


# don't rely on SingletonThreadPool
test_connection = ContextVar('test_connection', default=None)


class TestSessionFactory:

    def __init__(self):
        self._sm = sessionmaker()
        listen(self._sm, 'after_transaction_end', _restart_nested_transaction)

    def __call__(self) -> Session:
        conn = test_connection.get()
        if conn is not None:
            session = self._sm(bind=conn)
            session.begin_nested()
            return session
        else:
            raise RuntimeError('No test connection found in the current context')

    def dispose(self):
        pass


class DBTestMixin:

    def setUp(self, *args, **kwargs):
        super().setUp(*args, **kwargs)
        self._sf_patcher = patch('kios.factory.SessionFactory', new=TestSessionFactory)
        self._sf_patcher.start()
        old_conn = test_connection.get()
        if old_conn is not None:
            # if some tearDown got skipped (due to exception), clean up
            old_conn.close()
        conn = engine.connect()
        self._token = test_connection.set(conn)
        self._tr = conn.begin()

    def tearDown(self, *args, **kwargs):
        super().tearDown(*args, **kwargs)
        self._tr.rollback()
        conn = test_connection.get()
        test_connection.reset(self._token)
        conn.close()
        self._sf_patcher.stop()


@contextmanager
def temp_db_real_session_factory():
    with TemporaryDirectory() as d:
        p = Path(d) / 'test.db'
        with patch('kios.config.db_file', new=str(p)):
            sf = factory.SessionFactory()
            try:
                yield sf
            finally:
                sf.dispose()




#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from gettext import translation

t = translation('kios', fallback=True)
_ = t.gettext


class KiosError(Exception):
    pass


class DoRollback(Exception):
    pass


class UnsupportedAppPlatformError(KiosError):

    def __str__(self):
        return _('Unsupported application and platform')


class BusyDatabaseError(KiosError):

    def __str__(self):
        return _('Database is busy, please try again later')


class ExecutableAssignedError(KiosError):

    def __init__(self, app_name: str, exec_name):
        super().__init__(app_name, exec_name)
        self.app_name = app_name
        self.exec_name = exec_name

    def __str__(self):
        return _('Executable {exec_name} is associated with application {app_name}'.format(exec_name=self.exec_name,
                                                                                           app_name=self.app_name))


class ToolError(KiosError):

    def __init__(self, msg):
        super().__init__(msg)
        self.msg = msg

    def __str__(self):
        return _('Error running port scan application: {msg}').format(msg=self.msg)


class UnexpectedLineError(KiosError):

    def __init__(self, line, line_no):
        super().__init__(line, line_no)
        self.line = line
        self.line_no = line_no

    def __str__(self):
        return _('Unexpected line {line_no}: {line}').format(line=self.line, line_no=self.line_no)




#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from unittest import TestCase
from unittest.mock import patch

from sqlalchemy.exc import OperationalError

from kios.exception import BusyDatabaseError
from kios.data import TRANSPORT_PROTOCOL, NETWORK_PROTOCOL, PORT_NUMBER, EXECUTABLE_NAME, NetworkProtocol, \
    TransportProtocol
from kios.db import PersistenceManager, App, Executable, Port, present_session, session
from kios import factory

from .helper import DBTestMixin, temp_db_real_session_factory


class SessionManagerTestCase(DBTestMixin, TestCase):

    def test_commit(self):
        with factory.session_manager().database_session() as s:
            self.assertEqual(s.query(App).count(), 0)
            app = App(name='testapp')
            s.add(app)
        with factory.session_manager().database_session() as s:
            self.assertEqual(s.query(App).count(), 1)

    def test_rollback(self):
        with self.assertRaises(Exception):
            with factory.session_manager().database_session() as s:
                self.assertEqual(s.query(App).count(), 0)
                app = App(name='testapp')
                s.add(app)
                raise Exception
        with factory.session_manager().database_session() as s:
            self.assertEqual(s.query(App).count(), 0)

    def test_existing_session(self):
        with factory.session_manager().database_session():
            with self.assertRaises(RuntimeError):
                with factory.session_manager().database_session():
                    pass


class PresentSessionTestCase(TestCase):

    def test_no_present_session(self):
        with self.assertRaises(RuntimeError):
            with present_session():
                pass

    @patch('kios.config.db_busy_timeout', new=1)
    def test_error_handling(self):
        with temp_db_real_session_factory() as sf:
            s1 = sf()
            s2 = sf()
            token = session.set(s2)
            try:
                s1.add(App(name='app1'))
                s1.flush()
                with self.assertRaises(BusyDatabaseError):
                    with present_session() as s:
                        s.add(App(name='app2'))
                        s.flush()
                s2.rollback()
                with self.assertRaises(OperationalError):
                    with present_session() as s:
                        s.execute('ashdiaushd')
            finally:
                session.reset(token)
                s1.close()
                s2.close()


class PersistenceManagerTestCase(DBTestMixin, TestCase):

    @classmethod
    def setUpClass(cls):
        cls.pm = PersistenceManager()

    def setUp(self):
        super().setUp()
        with factory.session_manager().database_session() as s:
            app = App(name='testapp')
            app.executables = [Executable(name='testapp.exe'), Executable(name='testapp_helper.exe')]
            s.add(app)
            s.flush()
            self.app_id = app.id

    def test_saving_port_data(self):
        with factory.session_manager().database_session():
            port_data = [{TRANSPORT_PROTOCOL: TransportProtocol.UDP, NETWORK_PROTOCOL: NetworkProtocol.IPv4,
                          PORT_NUMBER: 1, EXECUTABLE_NAME: 'testapp.exe'},
                         {TRANSPORT_PROTOCOL: TransportProtocol.TCP, NETWORK_PROTOCOL: NetworkProtocol.IPv6,
                          PORT_NUMBER: 1, EXECUTABLE_NAME: 'testapp_helper.exe'}]
            self.pm.save_port_data(port_data)
        with factory.session_manager().database_session() as s:
            self.assertEqual(s.query(Port).count(), 2)
            self.assertEqual(s.query(Port).filter(Port.app_id == self.app_id, Port.number == 1,
                                                  Port.network_protocol == NetworkProtocol.IPv4,
                                                  Port.transport_protocol == TransportProtocol.UDP).count(), 1)
            self.assertEqual(s.query(Port).filter(Port.app_id == self.app_id, Port.number == 1,
                                                  Port.network_protocol == NetworkProtocol.IPv6,
                                                  Port.transport_protocol == TransportProtocol.TCP).count(), 1)

    def test_saving_port_data2(self):
        with self.assertRaises(Exception):
            with factory.session_manager().database_session():
                port_data = [{TRANSPORT_PROTOCOL: TransportProtocol.TCP, NETWORK_PROTOCOL: NetworkProtocol.IPv6,
                              PORT_NUMBER: 1, EXECUTABLE_NAME: 'testapp_helper.exe'}]
                self.pm.save_port_data(port_data, commit=True)
                raise Exception
        with factory.session_manager().database_session() as s:
            self.assertEqual(s.query(Port).count(), 1)
            self.assertEqual(s.query(Port).filter(Port.app_id == self.app_id, Port.number == 1,
                                                  Port.network_protocol == NetworkProtocol.IPv6,
                                                  Port.transport_protocol == TransportProtocol.TCP).count(), 1)



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from argparse import ArgumentParser
from enum import Enum, auto
from gettext import translation
from signal import signal, SIGINT
from sys import stdout, stderr
from typing import TYPE_CHECKING

from . import config
from .data import Platform, Application, NetworkProtocol, TransportProtocol
from .exception import DoRollback, KiosError
from .protocol import Feedback, OperationControl

if TYPE_CHECKING:
    from .core import KiosCore

t = translation('kios', fallback=True)
_ = t.gettext


class _Action(Enum):
    NOTHING = auto()
    SCAN = auto()
    IMPORT = auto()
    ADD = auto()
    PURGE = auto()
    LIST = auto()
    SHOW = auto()
    REMOVE = auto()


class ConsoleOperationControl(OperationControl):

    def __init__(self):
        self._interrupted = False

    def interrupt(self, *_) -> None:
        self._interrupted = True

    @property
    def interrupted(self) -> bool:
        return self._interrupted

    def feedback_pass_done(self):
        pass

    def __enter__(self):
        self._ph = signal(SIGINT, self.interrupt)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        signal(SIGINT, self._ph)
        if exc_val is not None and isinstance(exc_val, DoRollback):
            return True


def _info(data):
    print(data, file=stdout)


def _error(data):
    print(data, file=stderr)


class CLIFeedback(Feedback):

    def __init__(self):
        self.apps = []
        self.executables = []
        self.ports = []

    def app_entry(self, name: str):
        self.apps.append(name)

    def port_entry(self, np: NetworkProtocol, tp: TransportProtocol, number: int):
        self.ports.append((np, tp, number))

    def executable_entry(self, name: str):
        self.executables.append(name)

    def report_list(self):
        _info(_('Known applications:'))
        for name in self.apps:
            _info(f' {name}')

    def report_details(self):
        if len(self.apps) == 1:
            _info(_('Data for {app}').format(app=self.apps[0]))
            _info(_('Ports:'))
            for np, tp, n in self.ports:
                _info(f' {np.name}\t{tp.name}\t{n}')
            _info(_('Executables:'))
            for fn in self.executables:
                _info(f' {fn}')
        elif not self.apps:
            _info(_('No application found'))
        else:
            raise RuntimeError('Can\'t report details for multiple applications')


class CLI:

    def __init__(self, core: KiosCore):
        parser = ArgumentParser('kios')
        parser.set_defaults(action=_Action.NOTHING)
        parser.add_argument('--db_file', default=config.db_file,
                            help=_('path to a custom sqlite database file to use'))
        s_parsers = parser.add_subparsers()
        scan_parser = s_parsers.add_parser('scan', help=_('scan for ports applications listen at'))
        scan_parser.add_argument('interval', type=int, metavar='INTERVAL', nargs='?', default=config.port_scan_interval,
                                 help=_('time between scanner sweeps in seconds'))
        scan_parser.set_defaults(action=_Action.SCAN)
        import_parser = s_parsers.add_parser('import',
                                             help=_('process port data captured in a file'))
        import_parser.add_argument('file', metavar='FILE', help=_('path to a capture file'))
        import_parser.add_argument('--platform', choices=[item.name.lower() for item in Platform],
                                   default=config.platform.name.lower(),
                                   help=_('platform the output has been captured on'))
        import_parser.add_argument('--app', choices=[item.name.lower() for item in Application],
                                   default=config.app.name.lower(),
                                   help=_('application used to produce the captured output'))
        import_parser.set_defaults(action=_Action.IMPORT)
        add_parser = s_parsers.add_parser('add', help=_('add new application and executable or attach executable to an '
                                                        'existing app'))
        add_parser.add_argument('--force', default=False, action='store_true', help=_('reassign the executable to the '
                                                                                      'specified app'))
        add_parser.add_argument('app_name', metavar='APP_NAME', default=False, help=_('name of an application to '
                                                                                      '(create and) attach the '
                                                                                      'executable to'))
        add_parser.add_argument('exec_name', metavar='EXEC_NAME', default=False, help=_('name of an executable file to '
                                                                                        'associate with the '
                                                                                        'application'))
        add_parser.set_defaults(action=_Action.ADD)
        purge_parser = s_parsers.add_parser('purge', help=_('delete all information about application with a '
                                                            'specified name, including associated executables and '
                                                            'collected port data'))
        purge_parser.add_argument('app_name', metavar='APP_NAME', help=_('name of an application to purge'))
        purge_parser.set_defaults(action=_Action.PURGE)
        list_parser = s_parsers.add_parser('list', help=_('print names of known applications'))
        list_parser.set_defaults(action=_Action.LIST)
        show_parser = s_parsers.add_parser('show', help=_('print known ports and assigned executables for application '
                                                          'with a specified name'))
        show_parser.add_argument('app_name', metavar='APP_NAME', help=_('name of an application to print data on'))
        show_parser.set_defaults(action=_Action.SHOW)
        remove_parser = s_parsers.add_parser('remove', help=_('remove executable with a specified name'))
        remove_parser.add_argument('exec_name', metavar='EXECUTABLE_NAME', help=_('name of an executable to remove'))
        remove_parser.set_defaults(action=_Action.REMOVE)
        self._ns = parser.parse_args()
        self._core = core

    def run(self):
        config.db_file = self._ns.db_file
        try:
            act = self._ns.action
            if act == _Action.SCAN:
                config.port_scan_interval = self._ns.interval
                _info(_('Scanning system each {port_scan_interval} second(s), Ctrl+C to stop...')
                      .format(port_scan_interval=config.port_scan_interval))
                with ConsoleOperationControl() as control:
                    self._core.scan(control)
                _info(_('Done'))
            elif act == _Action.IMPORT:
                config.platform = Platform[self._ns.platform.upper()]
                config.app = Application[self._ns.app.upper()]
                config.data_file = self._ns.file
                _info(_('Importing {app} ({plat}) data from {df}, Ctrl+C to interrupt...')
                      .format(app=config.app.name.lower(), plat=config.platform.name.lower(), df=config.data_file))
                control = ConsoleOperationControl()
                with control:
                    self._core.import_(control)
                if control.interrupted:
                    _info(_('Interrupted'))
                else:
                    _info(_('Done'))
            elif act == _Action.ADD:
                self._core.add(self._ns.app_name, self._ns.exec_name, self._ns.force)
            elif act == _Action.PURGE:
                self._core.purge(self._ns.app_name)
            elif act == _Action.LIST:
                fb = CLIFeedback()
                self._core.list_(fb)
                fb.report_list()
            elif act == _Action.SHOW:
                fb = CLIFeedback()
                self._core.show_app(self._ns.app_name, fb)
                fb.report_details()
            elif act == _Action.REMOVE:
                self._core.remove_executable(self._ns.exec_name)
        except KiosError as e:
            _error(str(e))
            exit(1)



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from typing import TYPE_CHECKING
from sys import platform

from .cli import CLI
from .data import Platform, Application
from . import config, factory
from .operation import find_and_save_ports, save_ports_from_data_file

if TYPE_CHECKING:
    from .protocol import OperationControl, Feedback


class KiosCore:

    def __init__(self):
        self._sm = factory.session_manager()

    def scan(self, control: OperationControl):
        with self._sm.database_session():
            find_and_save_ports(control)

    def import_(self, control: OperationControl):
        with self._sm.database_session():
            save_ports_from_data_file(control)

    def add(self, app_name: str, exec_name: str, force: bool):
        with self._sm.database_session():
            factory.persistence_manager().add_executable_to_app(app_name, exec_name, force)

    def purge(self, app_name: str):
        with self._sm.database_session():
            factory.persistence_manager().purge_app(app_name)

    def list_(self, fb: Feedback):
        with self._sm.database_session():
            for app in factory.persistence_manager().list_apps():
                fb.app_entry(app.name)

    def show_app(self, name: str, fb: Feedback):
        with self._sm.database_session():
            app = factory.persistence_manager().get_app(name)
            if app is not None:
                fb.app_entry(app.name)
                for port in app.ports:
                    fb.port_entry(port.network_protocol, port.transport_protocol, port.number)
                for executable in app.executables:
                    fb.executable_entry(executable.name)

    def remove_executable(self, exec_name: str):
        with self._sm.database_session():
            factory.persistence_manager().remove_executable(exec_name)


def set_platform_defaults():
    if platform == 'win32':
        config.platform = Platform.WINDOWS
        config.app = Application.NETSTAT
        config.db_file = r'%LOCALAPPDATA%\kios\kios.db'


def main():
    set_platform_defaults()
    CLI(KiosCore()).run()



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING

from sqlalchemy import Column, Integer, String, Enum, ForeignKey, UniqueConstraint, select, bindparam
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlite3 import OperationalError as sq3_OperationalError

from .data import NetworkProtocol, TransportProtocol, NETWORK_PROTOCOL, TRANSPORT_PROTOCOL, PORT_NUMBER, EXECUTABLE_NAME
from .exception import BusyDatabaseError, ExecutableAssignedError

if TYPE_CHECKING:
    from typing import Iterable, Mapping, Any, Iterator, Optional

    from sqlalchemy.orm.session import Session

    from .factory import SessionFactory

Base = declarative_base()


class App(Base):
    __tablename__ = 'app'
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False, unique=True)


class Executable(Base):
    __tablename__ = 'executable'
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False, unique=True)
    app_id = Column(Integer, ForeignKey(App.id, ondelete='CASCADE'), nullable=False)
    app = relationship(App, back_populates='executables')


App.executables = relationship(Executable, back_populates='app', passive_deletes=True,
                               order_by=Executable.name)


class Port(Base):
    __tablename__ = 'port'
    __table_args__ = (UniqueConstraint('app_id', 'network_protocol', 'transport_protocol', 'number',
                                       sqlite_on_conflict='IGNORE'),)
    id = Column(Integer, primary_key=True)
    network_protocol = Column(Enum(NetworkProtocol), nullable=False)
    transport_protocol = Column(Enum(TransportProtocol), nullable=False)
    number = Column(Integer, nullable=False)
    app_id = Column(Integer, ForeignKey(App.id, ondelete='CASCADE'), nullable=False)
    app = relationship(App, back_populates='ports')


App.ports = relationship(Port, back_populates='app', passive_deletes=True,
                         order_by=Port.number & Port.transport_protocol & Port.network_protocol)

_port_t = Port.__table__
_executable_t = Executable.__table__

session: ContextVar[Session] = ContextVar('session', default=None)


class SessionManager:

    def __init__(self, session_factory: SessionFactory):
        self._sf = session_factory

    @contextmanager
    def database_session(self) -> Iterator[Session]:
        s = session.get()
        if s is None:
            s = None
            token = None
            try:
                s = self._sf()
                token = session.set(s)
                yield s
                s.commit()
            except Exception:
                s.rollback()
                raise
            finally:
                if token is not None:
                    session.reset(token)
                if s is not None:
                    s.close()
        else:
            raise RuntimeError('Session is already present in the current context')


@contextmanager
def present_session() -> Iterator[Session]:
    s = session.get()
    try:
        if s is not None:
            yield s
        else:
            raise RuntimeError('No database session in the current context')
    except OperationalError as e:
        if isinstance(e.orig, sq3_OperationalError):
            msg, = e.orig.args
            # best we can do with sqlite3
            if msg == 'database is locked':
                raise BusyDatabaseError from e
        raise


_app_id_select = select([bindparam(NETWORK_PROTOCOL, type_=Enum(NetworkProtocol)),
                         bindparam(TRANSPORT_PROTOCOL, type_=Enum(TransportProtocol)),
                         bindparam(PORT_NUMBER, type_=Integer),
                         _executable_t.c.app_id]).where(_executable_t.c.name == bindparam(EXECUTABLE_NAME))
_port_insert = _port_t.insert().from_select([_port_t.c.network_protocol, _port_t.c.transport_protocol,
                                             _port_t.c.number, _port_t.c.app_id], _app_id_select)


class PersistenceManager:

    @staticmethod
    def save_port_data(ports: Iterable[Mapping[str, Any]], *, commit=False) -> None:
        with present_session() as s:
            s.execute(_port_insert, ports)
            if commit:
                s.commit()

    @staticmethod
    def add_executable_to_app(app_name: str, exec_name: str, force: bool) -> None:
        with present_session() as s:
            app = s.query(App).filter(App.name == app_name).one_or_none()
            if app is None:
                app = App(name=app_name)
                s.add(app)
            executable = s.query(Executable).filter(Executable.name == exec_name).one_or_none()
            if executable is None:
                executable = Executable(name=exec_name)
                s.add(executable)
            elif executable.app is not app and not force:
                raise ExecutableAssignedError(executable.app.name, exec_name)
            executable.app = app

    @staticmethod
    def purge_app(name: str) -> None:
        with present_session() as s:
            s.query(App).filter(App.name == name).delete(synchronize_session=False)

    @staticmethod
    def list_apps() -> Iterable[App]:
        with present_session() as s:
            return s.query(App).order_by(App.name).all()

    @staticmethod
    def get_app(name: str) -> Optional[App]:
        with present_session() as s:
            return s.query(App).filter(App.name == name).one_or_none()

    @staticmethod
    def remove_executable(name: str) -> None:
        with present_session() as s:
            s.query(Executable).filter(Executable.name == name).delete(synchronize_session=False)




#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from subprocess import run, CalledProcessError

from .exception import ToolError


class WindowsNetstatDataSource:

    @staticmethod
    def get_port_data() -> str:
        res = None
        try:
            res = run(['netstat', '-a', '-n', '-b'], capture_output=True, text=True)
            res.check_returncode()
            return res.stdout
        except CalledProcessError as e:
            if e.returncode == 1 and res is not None:
                raise ToolError(res.stdout) from e
            else:
                raise



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from subprocess import CalledProcessError
from unittest import TestCase
from unittest.mock import patch

from kios.exception import ToolError
from kios.system import WindowsNetstatDataSource
from .helper import get_test_data


class WindowsNetstatDataSourceTestCase(TestCase):
    ds = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.ds = WindowsNetstatDataSource()

    @patch(target='kios.system.run', autospec=True)
    def test_getting_data(self, run):
        data = get_test_data('win_netstat.txt')
        type(run.return_value).stdout = data
        self.assertEqual(self.ds.get_port_data(), data)
        run.assert_called_once_with(['netstat', '-a', '-n', '-b'], capture_output=True, text=True)

    @patch(target='kios.system.run', autospec=True)
    def test_getting_data2(self, run):
        run.return_value.stdout = 'test'
        run.return_value.check_returncode.side_effect = CalledProcessError(1, [])
        with self.assertRaises(ToolError) as c:
            self.ds.get_port_data()
        self.assertEqual(c.exception.msg, 'test')

    @patch(target='kios.system.run', autospec=True)
    def test_getting_data3(self, run):
        run.return_value.check_returncode.side_effect = CalledProcessError(2, [])
        with self.assertRaises(Exception):
            self.ds.get_port_data()



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from functools import partial
from os.path import expandvars
from pathlib import Path
from typing import TYPE_CHECKING

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.event import listen

from . import config
from .data import Platform, Application
from .db import PersistenceManager, SessionManager, Base
from .exception import UnsupportedAppPlatformError
from .parser import WindowsNetstatPortDataParser
from .system import WindowsNetstatDataSource

if TYPE_CHECKING:
    from .parser import PortCollector


def data_source():
    if config.app == Application.NETSTAT and config.platform == Platform.WINDOWS:
        return WindowsNetstatDataSource()
    else:
        raise UnsupportedAppPlatformError


def data_parser(pc: PortCollector):
    if config.app == Application.NETSTAT and config.platform == Platform.WINDOWS:
        return WindowsNetstatPortDataParser(pc)
    else:
        raise UnsupportedAppPlatformError


def persistence_manager():
    return PersistenceManager()


def _errormaker():
    raise RuntimeError('Unable to create new session with disposed engine')


def _on_connect_handler(busy_timeout, conn, _):
    conn.isolation_level = None
    if config.db_debug:
        conn.set_trace_callback(print)
    cur = conn.cursor()
    try:
        cur.execute('PRAGMA foreign_keys = ON')
        cur.execute(f'PRAGMA busy_timeout = {busy_timeout}')
    finally:
        cur.close()


def _on_begin_handler(conn):
    conn.execute("BEGIN")


def engine(db_url: str):
    busy_timeout = config.db_busy_timeout
    if not isinstance(busy_timeout, int):
        raise RuntimeError(f'kios.config.db_busy_timeout has unexpected type {type(busy_timeout)}, int expected')
    e = create_engine(db_url, echo=config.db_debug, isolation_level=config.db_isolation)
    listen(e, "connect", partial(_on_connect_handler, busy_timeout))
    listen(e, "begin", _on_begin_handler)
    # TODO: look into embedding alembic <AP>
    Base.metadata.create_all(e)
    return e


class SessionFactory:

    def __init__(self):
        db_file = config.db_file
        if db_file is None:
            raise UnsupportedAppPlatformError
        p = Path(expandvars(db_file))
        p.parent.mkdir(parents=True, exist_ok=True)
        self.engine = engine(f'sqlite:///{str(p)}')
        self._sf = sessionmaker(bind=self.engine)

    def __call__(self):
        return self._sf()

    def dispose(self):
        self._sf = _errormaker
        self.engine.dispose()


def session_manager():
    return SessionManager(SessionFactory())



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from contextlib import contextmanager
from enum import Enum, auto
from ipaddress import ip_address
import re
from typing import TYPE_CHECKING

from transitions import Machine, MachineError

from .data import NETWORK_PROTOCOL, TRANSPORT_PROTOCOL, PORT_NUMBER, EXECUTABLE_NAME, NetworkProtocol, \
    TransportProtocol, ConnectionState
from .exception import UnexpectedLineError

if TYPE_CHECKING:
    from typing import Optional

_protocol = rf'(?P<proto>' + '|'.join(pr.name for pr in TransportProtocol) + ')'
_spacing = r'(?: +)'
_state = r'(?P<state>' + '|'.join(st.name for st in ConnectionState) + ')'
_hex_number = r'(?:(?:\d|[a-fA-F])+)'
_local_ip4_address = r'(?P<local_ipv4_address>\d+\.\d+\.\d+\.\d+)'
_foreign_ip4_address = r'(?:\d+\.\d+\.\d+\.\d+)'
_ip6_delimiter = r'(?:::|:)'
_ip6_zone = r'(?:%\d+)'
_local_ip6_address = rf'(?:\[(?P<local_ipv6_address>(?:{_hex_number}?{_ip6_delimiter})+{_hex_number}?)' \
                     rf'{_ip6_zone}?\])'
_foreign_ip6_address = rf'(?:\[(?:{_hex_number}?{_ip6_delimiter})+{_hex_number}?{_ip6_zone}?\])'
_local_port = r'(?P<local_port>\d+)'
_foreign_port = r'(?:\d+)'
_local_address = f'(?:(?:{_local_ip4_address}|{_local_ip6_address}):{_local_port})'
_wildcard_address = r'(?:\*:\*)'
_foreign_address = f'(?:(?:(?:{_foreign_ip4_address}|{_foreign_ip6_address}):{_foreign_port})|{_wildcard_address})'
_network_entry = rf'(?:^{_spacing}?{_protocol}{_spacing}{_local_address}{_spacing}{_foreign_address}' \
                 rf'(?:{_spacing}{_state}?)?$)'
_process_name = r'(?P<proc_name>.+)'
_process_entry = rf'(?:^{_spacing}?\[{_process_name}\]$)'
_generic_entry = rf'(?:^.+$)'
_data_entry = rf'(?:{_network_entry}|{_process_entry}|{_generic_entry})'
_data_entry_line_re = re.compile(_data_entry)


class PortCollector:

    def __init__(self):
        self.reset()

    def __call__(self, file_name: Optional[str], net_proto: NetworkProtocol, ip_addr: str,
                 transport_proto: TransportProtocol, port_number: int,
                 connection_state: Optional[ConnectionState]) -> None:
        ad = ip_address(ip_addr)
        if (connection_state == ConnectionState.LISTENING or transport_proto == TransportProtocol.UDP) \
                and not ad.is_loopback:
            self.port_data.append({EXECUTABLE_NAME: file_name, NETWORK_PROTOCOL: net_proto,
                                   TRANSPORT_PROTOCOL: transport_proto, PORT_NUMBER: port_number})

    def reset(self):
        self.port_data = []


class _NetstatState(Enum):
    INITIAL = auto()
    WAITING_NET = auto()
    WAITING_PROC = auto()
    DONE = auto()


class _NetstatModel:

    def __init__(self, pc: PortCollector):
        self._cb = pc
        self._reset()

    def on_exit_WAITING_PROC(self):
        self._cb(self.proc_name, self.network_protocol, self.ip_addr, self.transport_protocol, self.local_port,
                 self.conn_state)
        self._reset()

    def _reset(self):
        self.network_protocol = None
        self.ip_addr = None
        self.transport_protocol = None
        self.local_port = None
        self.conn_state = None
        self.proc_name = None


class WindowsNetstatPortDataParser:

    def __init__(self, pc: PortCollector):
        self._model = _NetstatModel(pc)
        states = [_NetstatState.INITIAL, _NetstatState.WAITING_NET, _NetstatState.WAITING_PROC, _NetstatState.DONE]
        transitions = [['start', [_NetstatState.INITIAL, _NetstatState.DONE], _NetstatState.WAITING_NET],
                       ['handle_net', [_NetstatState.WAITING_NET, _NetstatState.WAITING_PROC],
                        _NetstatState.WAITING_PROC],
                       ['handle_proc', _NetstatState.WAITING_PROC, _NetstatState.WAITING_NET],
                       ['finish', [_NetstatState.INITIAL, _NetstatState.WAITING_NET, _NetstatState.WAITING_PROC],
                        _NetstatState.DONE]]
        self._machine = Machine(model=self._model, states=states, transitions=transitions,
                                initial=_NetstatState.INITIAL)

    def _handle_data(self, proto, ipv4_addr, ipv6_addr, port, conn_state, proc_name):
        if proto is not None:
            self._model.handle_net()
            if ipv6_addr is None:
                self._model.network_protocol = NetworkProtocol.IPv4
                self._model.ip_addr = ipv4_addr
            else:
                self._model.network_protocol = NetworkProtocol.IPv6
                self._model.ip_addr = ipv6_addr
            self._model.transport_protocol = TransportProtocol[proto]
            self._model.local_port = int(port)
            try :
                self._model.conn_state = ConnectionState[conn_state] if conn_state is not None else None
            except KeyError:
                self._model.conn_state = None
        if proc_name is not None:
            self._model.proc_name = proc_name
            self._model.handle_proc()

    def extract_data(self, data: str):
        self._model.start()
        for line_no, line in enumerate(data.splitlines()):
            self._consume_line(line, line_no + 1)
        self._model.finish()

    def _consume_line(self, line: str, line_no: int):
        line = line.rstrip()
        match = _data_entry_line_re.match(line)
        if match:
            proto, ipv4_addr, ipv6_addr, port, conn_state, proc_name = match.groups()
            try:
                self._handle_data(proto, ipv4_addr, ipv6_addr, port, conn_state, proc_name)
            except MachineError:
                raise UnexpectedLineError(line, line_no)

    @contextmanager
    def line_stream_consumer(self):
        self._model.start()
        yield self._consume_line
        self._model.finish()



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from signal import SIGINT

from unittest import TestCase
from unittest.mock import patch, NonCallableMock, DEFAULT

from kios.cli import ConsoleOperationControl, CLIFeedback, CLI, _info, _error
from kios import config
from kios.core import KiosCore
from kios.data import NetworkProtocol, TransportProtocol
from kios.exception import DoRollback, ExecutableAssignedError, ToolError, BusyDatabaseError, \
    UnsupportedAppPlatformError, UnexpectedLineError
from .helper import patch_config_app_platform, TestConsoleOperationControl, inside_test_console_operation_control


class OperationControlTestCase(TestCase):

    def setUp(self) -> None:
        self.oc = ConsoleOperationControl()

    def test_interruption(self):
        self.assertFalse(self.oc.interrupted)
        self.oc.interrupt(1, 2)
        self.assertTrue(self.oc.interrupted)

    @patch('kios.cli.signal', autospec=True)
    def test_signal_management(self, signal):
        with self.oc:
            signal.assert_called_once_with(SIGINT, self.oc.interrupt)
            prev = signal.return_value
            signal.reset_mock()
        signal.assert_called_once_with(SIGINT, prev)

    def test_exception_handling(self):
        with self.oc:
            raise DoRollback

    def test_exception_handling2(self):
        with self.assertRaises(Exception):
            with self.oc:
                raise Exception

    def test_feedback_pass(self):
        self.oc.feedback_pass_done()


class CLIFeedbackTestCase(TestCase):

    def setUp(self) -> None:
        self.f = CLIFeedback()

    @patch('kios.cli._info', autospec=True)
    def test_reporting_app_list(self, _info):
        self.f.app_entry('app1')
        self.f.app_entry('app2')
        self.f.report_list()
        _info.assert_called()

    @patch('kios.cli._info', autospec=True)
    def test_reporting_app_details(self, _info):
        self.f.app_entry('app1')
        self.f.port_entry(NetworkProtocol.IPv4, TransportProtocol.TCP, 1)
        self.f.executable_entry('app1.exe')
        self.f.report_details()
        _info.assert_called()

    def test_reporting_app_details2(self):
        with self.assertRaises(RuntimeError):
            self.f.app_entry('app1')
            self.f.app_entry('app2')
            self.f.report_details()

    @patch('kios.cli._info', autospec=True)
    def test_reporting_app_details3(self, _info):
        self.f.report_details()
        _info.assert_called()


def patch_output(t):
    pex = patch('builtins.exit', autospec=True)
    pio = patch.multiple('kios.cli', _info=DEFAULT, _error=DEFAULT, autospec=True)
    return pex(pio(t))


class _CoreMixin:

    def setUp(self) -> None:
        self.core = NonCallableMock(spec_set=KiosCore)


@patch_config_app_platform
@patch('kios.cli.ConsoleOperationControl', new=TestConsoleOperationControl)
@patch_output
class CLIOperationTestCase(_CoreMixin, TestCase):

    @patch('kios.config.port_scan_interval', new=None)
    @patch('sys.argv', new=['kios', 'scan', '15'])
    def test_scanning(self, exit, _info, _error):

        def check_inside_context(oc):
            inside_test_console_operation_control(oc)
            self.assertEqual(config.port_scan_interval, 15)

        self.core.scan.side_effect = check_inside_context
        CLI(self.core).run()
        self.core.scan.assert_called_once()
        _info.assert_called()
        _error.assert_not_called()
        exit.assert_not_called()

    @patch('sys.argv', new=['kios', 'scan'])
    def test_scanning2(self, exit, _info, _error):
        self.core.scan.side_effect = ToolError('abc')
        CLI(self.core).run()
        self.core.scan.assert_called_once()
        _info.assert_called()
        _error.assert_called()
        exit.assert_called_once_with(1)

    @patch('kios.config.data_file', new=None)
    @patch('sys.argv', new=['kios', 'import', r'abc.txt'])
    def test_importing(self, exit, _info, _error):

        def check_inside_context(oc):
            inside_test_console_operation_control(oc)
            self.assertEqual(config.data_file, 'abc.txt')

        self.core.import_.side_effect = check_inside_context
        CLI(self.core).run()
        self.core.import_.assert_called_once()
        _info.assert_called()
        _error.assert_not_called()
        exit.assert_not_called()

    @patch('kios.config.data_file', new=None)
    @patch('sys.argv', new=['kios', 'import', r'abc.txt'])
    def test_importing2(self, exit, _info, _error):
        def check_inside_context(oc):
            inside_test_console_operation_control(oc)
            self.assertEqual(config.data_file, 'abc.txt')
            oc.interrupt()

        self.core.import_.side_effect = check_inside_context
        CLI(self.core).run()
        self.core.import_.assert_called_once()
        _info.assert_called()
        _error.assert_not_called()
        exit.assert_not_called()


@patch_config_app_platform
@patch_output
class ManagementTestCase(_CoreMixin, TestCase):

    @patch('sys.argv', new=['kios', 'add', 'app1', 'app.exe'])
    def test_adding_apps(self, exit, _info, _error):
        CLI(self.core).run()
        self.core.add.assert_called_once_with('app1', 'app.exe', False)
        _info.assert_not_called()
        _error.assert_not_called()
        exit.assert_not_called()

    @patch('sys.argv', new=['kios', 'add', 'app2', 'app.exe'])
    def test_adding_apps2(self, exit, _info, _error):
        self.core.add.side_effect = ExecutableAssignedError('app2', 'app.exe')
        CLI(self.core).run()
        self.core.add.assert_called_once_with('app2', 'app.exe', False)
        _info.assert_not_called()
        _error.assert_called_once()
        exit.assert_called_once_with(1)

    @patch('sys.argv', new=['kios', 'add', '--force', 'app2', 'app.exe'])
    def test_adding_apps3(self, exit, _info, _error):
        CLI(self.core).run()
        self.core.add.assert_called_once_with('app2', 'app.exe', True)
        _info.assert_not_called()
        _error.assert_not_called()
        exit.assert_not_called()

    @patch('sys.argv', new=['kios', 'purge', 'app1'])
    def test_purging_apps(self, exit, _info, _error):
        CLI(self.core).run()
        self.core.purge.assert_called_once_with('app1')
        _info.assert_not_called()
        _error.assert_not_called()
        exit.assert_not_called()

    @patch('kios.cli.CLIFeedback', autospec=True)
    @patch('sys.argv', new=['kios', 'list'])
    def test_listing_apps(self, CLIFeedback, exit, _info, _error):
        CLI(self.core).run()
        fb = CLIFeedback.return_value
        self.core.list_.assert_called_once_with(fb)
        fb.report_list.assert_called_once()
        _error.assert_not_called()
        exit.assert_not_called()

    @patch('kios.cli.CLIFeedback', autospec=True)
    @patch('sys.argv', new=['kios', 'show', 'app'])
    def test_showing_details(self, CLIFeedback, exit, _info, _error):
        CLI(self.core).run()
        fb = CLIFeedback.return_value
        self.core.show_app.assert_called_once_with('app', fb)
        fb.report_details.assert_called_once()
        _error.assert_not_called()
        exit.assert_not_called()

    @patch('sys.argv', new=['kios', 'remove', 'app.exe'])
    def test_removing_executables(self, exit, _info, _error):
        CLI(self.core).run()
        self.core.remove_executable.assert_called_once_with('app.exe')
        _info.assert_not_called()
        _error.assert_not_called()
        exit.assert_not_called()


class SkippedExceptionMessageTestCase(TestCase):

    def test_busy_db_message(self):
        str(BusyDatabaseError())

    def test_unsupported_app_platform_message(self):
        str(UnsupportedAppPlatformError())

    def test_unexpected_line_message(self):
        str(UnexpectedLineError('abc', 1))


@patch('builtins.print', autospec=True)
@patch.multiple('kios.cli', stdout=DEFAULT, stderr=DEFAULT, autospec=True)
class OutputTestCase(TestCase):

    def test_info(self, print, stdout, stderr):
        _info('abc')
        print.assert_called_once_with('abc', file=stdout)

    def test_error(self, print, stdout, stderr):
        _error('abc')
        print.assert_called_once_with('abc', file=stderr)



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from enum import Enum, auto

NETWORK_PROTOCOL = 'network_protocol'
TRANSPORT_PROTOCOL = 'transport_protocol'
PORT_NUMBER = 'port_number'
EXECUTABLE_NAME = 'executable_name'


class NetworkProtocol(Enum):
    IPv4 = auto()
    IPv6 = auto()


class TransportProtocol(Enum):
    UDP = auto()
    TCP = auto()


class ConnectionState(Enum):
    LISTENING = auto()
    ESTABLISHED = auto()
    SYN_SENT = auto()
    CLOSE_WAIT = auto()
    TIME_WAIT = auto()
    CLOSED = auto()
    FIN_WAIT_1 = auto()
    FIN_WAIT_2 = auto()
    LAST_ACK = auto()
    SYN_RECEIVED = auto()


class Platform(Enum):
    WINDOWS = auto()


class Application(Enum):
    NETSTAT = auto()



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from time import sleep, time
from typing import TYPE_CHECKING

from . import config, factory
from .exception import DoRollback
from .parser import PortCollector

if TYPE_CHECKING:
    from .protocol import OperationControl


def find_and_save_ports(control: OperationControl) -> None:
    ds = factory.data_source()
    pc = PortCollector()
    dp = factory.data_parser(pc)
    pm = factory.persistence_manager()
    while not control.interrupted:
        dp.extract_data(ds.get_port_data())
        if pc.port_data:
            pm.save_port_data(pc.port_data, commit=True)
            pc.reset()
        scan_done_time = time()
        control.feedback_pass_done()
        while not control.interrupted and time() - scan_done_time < config.port_scan_interval:
            sleep(config.port_scan_wake_up_interval)


def save_ports_from_data_file(control: OperationControl) -> None:
    sf = config.data_file
    if sf is not None:
        pc = PortCollector()
        dp = factory.data_parser(pc)
        pm = factory.persistence_manager()
        bs = config.import_batch_size
        with open(sf, 'rt') as fo:
            with dp.line_stream_consumer() as consume:
                for line_no, line in enumerate(fo):
                    if control.interrupted:
                        raise DoRollback
                    consume(line, line_no + 1)
                    if len(pc.port_data) >= bs:
                        pm.save_port_data(pc.port_data)
                        pc.reset()
                        control.feedback_pass_done()
                if pc.port_data:
                    pm.save_port_data(pc.port_data)
                    control.feedback_pass_done()
    else:
        raise RuntimeError('No data file is specified in the config')



#
# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2020 Andrey Pleshakov
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from unittest import TestCase
from unittest.mock import patch, NonCallableMock, call, DEFAULT, Mock

from kios import factory
from kios.core import KiosCore, set_platform_defaults, main
from kios import config
from kios.data import NetworkProtocol, TransportProtocol, Application, Platform
from kios.db import App, Executable, Port
from kios.exception import DoRollback, ExecutableAssignedError
from kios.protocol import Feedback
from .helper import DBTestMixin, get_test_data, TestOperationControl, path_config_intervals, get_test_file_path, \
    patch_config_app_platform

_data_se = [get_test_data('win_netstat2.txt'), get_test_data('win_netstat2.txt')]

patch_data_file = patch('kios.config.data_file', new=get_test_file_path('win_netstat.txt'))


@patch_config_app_platform
class PortActionsTestCase(DBTestMixin, TestCase):

    def setUp(self):
        super().setUp()
        with factory.session_manager().database_session() as s:
            app1 = App(name='app1')
            app1.executables = [Executable(name='app1.exe')]
            app2 = App(name='app2')
            app2.executables = [Executable(name='app2.exe')]
            app3 = App(name='app3')
            app3.executables = [Executable(name='app3.exe')]
            s.add_all([app1, app2, app3])
        self.core = KiosCore()

    @path_config_intervals
    @patch('kios.factory.data_source', autospec=True, **{'return_value.get_port_data.side_effect': _data_se * 2})
    def test_scanning(self, data_source):
        self.core.scan(TestOperationControl())
        self.verify_state()
        self.core.scan(TestOperationControl())
        self.verify_state()

    @patch('kios.config.data_file', new=get_test_file_path('win_netstat2.txt'))
    def test_importing(self):
        self.core.import_(TestOperationControl())
        self.verify_state()
        self.core.import_(TestOperationControl())
        self.verify_state()

    @patch_data_file
    def test_importing2(self):
        self.core.import_(TestOperationControl())
        with factory.session_manager().database_session() as s:
            self.assertEqual(s.query(Port).count(), 4)

    @patch_data_file
    @patch('kios.config.import_batch_size', new=2)
    def test_importing3(self):
        with self.assertRaises(DoRollback):
            self.core.import_(TestOperationControl(1))
        with factory.session_manager().database_session() as s:
            self.assertEqual(s.query(Port).count(), 0)

    def verify_state(self):
        with factory.session_manager().database_session() as s:
            self.assertEqual(s.query(Port).join(App).join(Executable)
                             .filter(Executable.name == 'app3.exe', Port.network_protocol == NetworkProtocol.IPv6,
                                     Port.transport_protocol == TransportProtocol.UDP, Port.number == 3).count(), 1)


class AppActionsTestCase(DBTestMixin, TestCase):

    def setUp(self):
        super().setUp()
        self.core = KiosCore()

    def test_addition(self):
        self.core.add('testapp1', 'testapp.exe', False)
        self.verify_state()
        self.core.add('testapp1', 'testapp.exe', False)
        self.verify_state()

    def test_addition2(self):
        self.core.add('testapp1', 'testapp.exe', False)
        self.verify_state()
        with self.assertRaises(ExecutableAssignedError) as c:
            self.core.add('testapp2', 'testapp.exe', False)
        self.assertEqual(c.exception.app_name, 'testapp1')
        self.assertEqual(c.exception.exec_name, 'testapp.exe')
        self.verify_state()

    def test_addition3(self):
        self.core.add('testapp1', 'testapp.exe', False)
        self.verify_state()
        self.core.add('testapp2', 'testapp.exe', True)
        with factory.session_manager().database_session() as s:
            self.assertEqual(s.query(App).count(), 2)
            self.assertEqual(s.query(App).filter(App.name == 'testapp1').count(), 1)
            app2 = s.query(App).filter(App.name == 'testapp2').one()
            self.assertEqual(s.query(Executable).count(), 1)
            self.assertEqual(s.query(Executable).filter(Executable.name == 'testapp.exe',
                                                        Executable.app == app2).count(), 1)

    def verify_state(self):
        with factory.session_manager().database_session() as s:
            self.assertEqual(s.query(App).count(), 1)
            app = s.query(App).filter(App.name == 'testapp1').one()
            self.assertEqual(s.query(Executable).count(), 1)
            self.assertEqual(s.query(Executable).filter(Executable.name == 'testapp.exe',
                                                        Executable.app == app).count(), 1)


class AppActionsTestCase2(DBTestMixin, TestCase):

    def setUp(self):
        super().setUp()
        with factory.session_manager().database_session() as s:
            app = App(name='test1')
            app2 = App(name='test2')
            app.ports = [Port(network_protocol=NetworkProtocol.IPv4, transport_protocol=TransportProtocol.UDP,
                              number=1),
                         Port(network_protocol=NetworkProtocol.IPv6, transport_protocol=TransportProtocol.TCP,
                              number=2)]
            app.executables = [Executable(name='test1.exe'), Executable(name='test1_helper.exe')]
            s.add(app)
            app2.ports = [Port(network_protocol=NetworkProtocol.IPv6, transport_protocol=TransportProtocol.UDP,
                               number=3),
                          Port(network_protocol=NetworkProtocol.IPv4, transport_protocol=TransportProtocol.TCP,
                               number=4)]
            app2.executables = [Executable(name='test2.exe'), Executable(name='test2_helper.exe')]
            s.add(app2)
        self.core = KiosCore()

    def test_purging(self):
        self.core.purge('test1')
        with factory.session_manager().database_session() as s:
            self.assertEqual(s.query(App).count(), 1)
            self.assertEqual(s.query(Executable).count(), 2)
            self.assertEqual(s.query(Port).count(), 2)
            app2 = s.query(App).filter(App.name == 'test2').one()
            self.assertEqual(s.query(Executable).filter(Executable.name == 'test2.exe',
                                                        Executable.app == app2).count(), 1)
            self.assertEqual(s.query(Executable).filter(Executable.name == 'test2_helper.exe',
                                                        Executable.app == app2).count(), 1)
            self.assertEqual(s.query(Port).filter(Port.network_protocol == NetworkProtocol.IPv6,
                                                  Port.transport_protocol == TransportProtocol.UDP,
                                                  Port.number == 3, Port.app == app2).count(), 1)
            self.assertEqual(s.query(Port).filter(Port.network_protocol == NetworkProtocol.IPv4,
                                                  Port.transport_protocol == TransportProtocol.TCP,
                                                  Port.number == 4, Port.app == app2).count(), 1)

    def test_listing_apps(self):
        fb = NonCallableMock(spec_set=Feedback)
        self.core.list_(fb)
        self.assertEqual(fb.app_entry.call_count, 2)
        fb.app_entry.assert_has_calls([call('test1'), call('test2')])

    def test_showing_app(self):
        fb = NonCallableMock(spec_set=Feedback)
        self.core.show_app('test1', fb)
        fb.app_entry.called_once_with('test1')
        self.assertEqual(fb.port_entry.call_count, 2)
        fb.port_entry.assert_has_calls([call(NetworkProtocol.IPv4, TransportProtocol.UDP, 1),
                                        call(NetworkProtocol.IPv6, TransportProtocol.TCP, 2)], any_order=True)
        self.assertEqual(fb.executable_entry.call_count, 2)
        fb.executable_entry.assert_has_calls([call('test1.exe'), call('test1_helper.exe')], any_order=True)

    def test_showing_app2(self):
        fb = NonCallableMock(spec_set=Feedback)
        self.core.show_app('test3', fb)
        self.assertEqual(len(fb.method_calls), 0)

    def test_removing(self):
        self.core.remove_executable('test1_helper.exe')
        with factory.session_manager().database_session() as s:
            self.assertEqual(s.query(App).count(), 2)
            self.assertEqual(s.query(Executable).count(), 3)
            self.assertEqual(s.query(Port).count(), 4)
            app = s.query(App).filter(App.name == 'test1').one()
            self.assertEqual(s.query(Executable).filter(Executable.app == app).count(), 1)
            self.assertEqual(s.query(Executable).filter(Executable.name == 'test1.exe',
                                                        Executable.app == app).count(), 1)


class OtherTestCase(TestCase):

    @patch.multiple(config, platform=None, app=None)
    @patch('kios.core.platform', new='win32')
    def test_setting_platform_defaults(self):
        set_platform_defaults()
        self.assertEqual(config.app, Application.NETSTAT)
        self.assertEqual(config.platform, Platform.WINDOWS)

    @patch.multiple('kios.core', set_platform_defaults=DEFAULT, CLI=DEFAULT, KiosCore=DEFAULT, autospec=True)
    def test_main(self, set_platform_defaults, CLI, KiosCore):
        o = Mock()
        o.attach_mock(set_platform_defaults, 'spd')
        o.attach_mock(CLI, 'cli')
        o.attach_mock(KiosCore, 'kc')
        main()
        self.assertSequenceEqual(o.mock_calls, [call.spd(), call.kc(),
                                                *call.cli(KiosCore.return_value).run().call_list()])

